#ifndef __SIGLEX_HH__
#define __SIGLEX_HH__

#include <gecode/int.hh>

namespace SigLex {
    using namespace Gecode;
    void
    siglex(Home home, int k, IntArgs p, const IntVarArgs& x, IntVarArray *d=0, IntVarArray *y=0, IntVarArray *finp=0);

    void
    siglexdec(Home home, int k, IntArgs p, const IntVarArgs& x, IntVarArray *d=0, IntVarArray *y=0, IntVarArray *finp=0);
}

#include "siglex.hpp"

#endif
